////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file randomizer.c
/// Definitions of randomizer functions.
/// \example randomizer01.c
/// \example randomizer02.c
/// \example randomizer03.c
////////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/randomizer.h"
#include "../include/tscope5/randomizer_internal.h"
#include "../include/tscope5/system_internal.h"

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <sys/time.h>


////////////////////////////////////////////////////////////////////////////////
/// @name Random integers
/// In an experiment most of the time you will need random integers,
/// preferably drawn from a list without replacement:
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Draw an integer from a uniform distribution.
///
/// \param max Number of possible values of the random integer.
///
/// \return An integer from 0 to n-1.
////////////////////////////////////////////////////////////////////////////////
unsigned long int ts5_random_integer(const unsigned long int max)
{
	ts5_check_randomizer("ts5_random_integer");
	ts5_log(TS5_LOGLEVEL_5, "ts5_random_integer(%lu)\n", max);

	return gsl_rng_uniform_int(_ts5_data.randomizer, max);
}


////////////////////////////////////////////////////////////////////////////////
/// Draw a list of random values without replacement.
///
/// \param nmax Maximum value (output values will range from 0 to n-1).
/// \param freq Frequency of each value.
/// \param list Adress of the array where the list is stored.
////////////////////////////////////////////////////////////////////////////////
void ts5_random_list(const int nmax, const int freq, int *list)
{
	ts5_check_randomizer("ts5_random_list");
	ts5_log(TS5_LOGLEVEL_5, "ts5_random_list(%d,%d,%p)\n", nmax, freq, list);

	int i;
	for (i = 0; i < nmax * freq; i++) {
		list[i] = i % nmax;
	}

	gsl_ran_shuffle(_ts5_data.randomizer, list, (unsigned long)(nmax * freq), 
		sizeof(int));
}

////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// @name Random doubles
/// Tscope5 also gives acces to random numbers drawn from 
/// some common distributions:
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Draw a double from a uniform distribution.
///
/// \return A double precision floating point value from a uniform distribution.
///
/// The range includes 0.0 but excludes 1.0.
////////////////////////////////////////////////////////////////////////////////
double ts5_random_uniform()
{
	ts5_check_randomizer("ts5_random_uniform");
	ts5_log(TS5_LOGLEVEL_5, "ts5_random_uniform()\n");

	return gsl_rng_uniform(_ts5_data.randomizer);
}


////////////////////////////////////////////////////////////////////////////////
/// Draw a double from a normal distribution.
///
/// \param mu Mean of the distribution.
/// \param sigma Standard deviation of the distribution.
///
/// \return The random value.
////////////////////////////////////////////////////////////////////////////////
double ts5_random_normal(double mu, double sigma)
{
	ts5_check_randomizer("ts5_random_normal");
	ts5_log(TS5_LOGLEVEL_5, "ts5_random_normal(%f,%f)\n", mu, sigma);

	return (mu + gsl_ran_gaussian(_ts5_data.randomizer, sigma));
}


////////////////////////////////////////////////////////////////////////////////
/// Draw a double from an exponential distribution.
///
/// \param mu Mean of the distribution.
///
/// \return The random value.
////////////////////////////////////////////////////////////////////////////////
double ts5_random_exponential(double mu)
{
	ts5_check_randomizer("ts5_random_exponential");
	ts5_log(TS5_LOGLEVEL_5, "ts5_random_exponential(%f)\n", mu);

	return gsl_ran_exponential(_ts5_data.randomizer, mu);
}


////////////////////////////////////////////////////////////////////////////////
/// @name Set the random seed
/// If you do simulation work it is also possible to set 
/// the random seed yourself.
/// Identical seeds always result in the same randomized list.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Seed the randomizer subsystem.
///
/// \param seed Seed for the randomizer.
///
/// If the generator is seeded with the same value of seed 
/// on two different runs, the same stream of random numbers will be generated
/// by successive calls to the routines.
///
/// The timer is a good source for random seeds.
////////////////////////////////////////////////////////////////////////////////
unsigned long int ts5_seed_randomizer(const unsigned long int seed)
{
	ts5_check_randomizer("ts5_seed_randomizer");
	ts5_log(TS5_LOGLEVEL_4, "ts5_seed_randomizer(%lu)\n", seed);

	gsl_rng_set(_ts5_data.randomizer, seed);

	return seed;
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////
