////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file randomizer_internal.c
/// Definitions of internal randomizer functions.
////////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/randomizer_internal.h"
#include "../include/tscope5/system_internal.h"

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <sys/time.h>

/// Is the randomizer subsystem installed?
int _ts5_is_randomizer_installed = 0;


////////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each randomizer function.
///
/// \param calling_function Name the function that calls 
/// for this check or installation.
///
/// Checks whether the randomizer subsystem  is installed. 
/// If not, the randomizer subsystem is installed.
////////////////////////////////////////////////////////////////////////////////
void ts5_check_randomizer(char *calling_function)
{
	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_randomizer\n", calling_function);
	if (!_ts5_is_randomizer_installed) {
		ts5_install_randomizer(calling_function);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Install the randomizer subsystem.
///
/// \param calling_function Name the function that calls 
/// for this check or installation.
///
/// The ranlux randomizer from the Gnu Scientific Library is used.
///
/// This function is called automatically if necessary.
////////////////////////////////////////////////////////////////////////////////
void ts5_install_randomizer(char *calling_function)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5(calling_function);
	}

	if (!_ts5_is_randomizer_installed) {

		_ts5_is_randomizer_installed = 1;
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 randomizer\n", 
				calling_function);

		_ts5_data.randomizer = gsl_rng_alloc(gsl_rng_ranlxs2);

		if (!_ts5_data.randomizer) {
			ts5_fatal("%s: could not install Tscope5 randomizer\n", 
					calling_function);
		}

		// seed randomizer using the clock time
		struct timeval tv;
		gettimeofday(&tv, NULL);
		unsigned long seed = tv.tv_usec;
		gsl_rng_set(_ts5_data.randomizer, seed);
		ts5_log(TS5_LOGLEVEL_1, "%s: Random seed is %lu\n", 
				calling_function, seed);

		atexit(ts5_uninstall_randomizer);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Uninstall the randomizer subsystem.
///
/// This function is called automatically at the end of the program.
////////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_randomizer()
{
	if (_ts5_is_randomizer_installed) {

		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 randomizer\n");

		gsl_rng_free(_ts5_data.randomizer);
		_ts5_data.randomizer = NULL;

		_ts5_is_randomizer_installed = 0;
	}
}

