////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file serialport_internal.c
/// Definitions of internal serial port functions.
////////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/serialport_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/timer_internal.h"

#ifndef TS5_WINDOWS
	#include <sys/types.h>
	#include <sys/ioctl.h>
	#include <dirent.h>
	#include <termios.h>
	#include <errno.h>

#endif

/// Is the serialport subsystem installed?
int _ts5_is_serialport_installed = 0;


////////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each serialport function.
///
/// \param calling_function Name the function that calls 
/// for this check or installation.
///
/// Checks whether the serialport subsystem is installed. 
/// If not, the serialport subsystem is installed.
////////////////////////////////////////////////////////////////////////////////
void ts5_check_serialport(char *calling_function)
{
	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_serialport\n", calling_function);
	if (!_ts5_is_serialport_installed) {
		ts5_install_serialport(calling_function);
	}	
}


////////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each serialport function.
///
/// \param calling_function Name the function that calls 
/// for this check or installation.
/// \param portname Name of the port.
///
/// Checks whether this serial port is already opened. 
/// If so, return the port number. 
/// If not, open it and return the port number. 
////////////////////////////////////////////////////////////////////////////////
int ts5_check_serialport2(char *calling_function, char *portname)
{
	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_serialport2(%s)\n", 
			calling_function, portname);
	
	int i;
	int portnum = -1;
	
	// is the port already open? if so return the port index
	for (i=0; i<_ts5_status.timer.num_serialports; i++) {
		if (!strcmp(portname, _ts5_status.timer.serialport[i].portname)) {
			portnum = i;
		} 		
	}	
		
	// if the port is not open try to open it
	if (portnum == -1) {

		TS5_SERIALPORT *port;
		port = (TS5_SERIALPORT *)al_malloc(sizeof(TS5_SERIALPORT));
		
		#ifdef TS5_WINDOWS
			*port = CreateFile(portname,
						GENERIC_READ | GENERIC_WRITE,
						0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
						
						// asynchronous:
						// replace FILE_ATTRIBUTE_NORMAL
						// by FILE_FLAG_OVERLAPPED
						// http://msdn.microsoft.com/en-us/library/aa365683(VS.85).aspx
						
			if (*port == INVALID_HANDLE_VALUE) {
				ts5_fatal("%s: could not open serial port %s\n", 
					calling_function, portname);
			}
										
		#else
			*port = open(portname, O_RDWR | O_NOCTTY | O_NDELAY);

			if (*port == -1) {
				ts5_fatal("%s: could not open serial port %s\n", 
					calling_function, portname);
			}

			// get exclusive io
			if (ioctl(*port, TIOCEXCL) == -1) {
				ts5_fatal("%s: could not get exclusivity on serial port %s\n",
					calling_function, portname);
			}		
		#endif
						
		// if we get here we apparently managed to open the port
		_ts5_status.timer.num_serialports++;
		portnum = _ts5_status.timer.num_serialports - 1;

		// update status pointer
		TS5_SERIALPORT_STATUS *serialport_status;
		serialport_status = al_realloc(_ts5_status.timer.serialport, 
			_ts5_status.timer.num_serialports * sizeof(TS5_SERIALPORT_STATUS));

		if (serialport_status == NULL) {
			ts5_fatal("%s: %s\n", calling_function, 
				"failed to allocate memory for serial port settings");
		}

		_ts5_status.timer.serialport = 
			(TS5_SERIALPORT_STATUS *) serialport_status;
				
		// fill in status values
		strcpy(_ts5_status.timer.serialport[portnum].portname, portname);
			_ts5_status.timer.serialport[portnum].port = port;
		
		_ts5_status.timer.serialport[portnum].num_buttons = 127;
		_ts5_status.timer.serialport[portnum].num_defined_buttons = 0;
		_ts5_status.timer.serialport[portnum].num_active_buttons = 0;
		
		_ts5_status.timer.serialport[portnum].button_press_defined =
			(int *)al_malloc(sizeof(int)
			* _ts5_status.timer.serialport[portnum].num_buttons);

		_ts5_status.timer.serialport[portnum].button_press_active =
			(int *)al_malloc(sizeof(int)
			* _ts5_status.timer.serialport[portnum].num_buttons);

		_ts5_status.timer.serialport[portnum].button_release_defined =
			(int *)al_malloc(sizeof(int)
			* _ts5_status.timer.serialport[portnum].num_buttons);

		_ts5_status.timer.serialport[portnum].button_release_active =
			(int *)al_malloc(sizeof(int)
			* _ts5_status.timer.serialport[portnum].num_buttons);
			
		int j;
		for (j=0; j<_ts5_status.timer.serialport[portnum].num_buttons; j++) {
			_ts5_status.timer.serialport[portnum].button_press_defined[j] = 0;
			_ts5_status.timer.serialport[portnum].button_press_active[j] = 0;
			_ts5_status.timer.serialport[portnum].button_release_defined[j] = 0;
			_ts5_status.timer.serialport[portnum].button_release_active[j] = 0;
		}
				
		_ts5_status.timer.serialport[portnum].is_trigger_input_device = 0;
		_ts5_status.timer.serialport[portnum].is_trigger_output_device = 0;
		_ts5_status.timer.serialport[portnum].simulate_trigger_input = 0;
		_ts5_status.timer.serialport[portnum].trigger_simulation_interval = 2.0;
		_ts5_status.timer.serialport[portnum].last_trigger_simulation = 
			al_get_time();

	}
	
	
	return portnum;
}


////////////////////////////////////////////////////////////////////////////////
/// Install the serialport subsystem.
///
/// \param calling_function Name the function that calls 
/// for this check or installation.
///
/// This function is called automatically if necessary.
////////////////////////////////////////////////////////////////////////////////
void ts5_install_serialport(char *calling_function)
{
	if (!_ts5_is_timer_installed) {
		ts5_install_timer(calling_function);
	}

	if (!_ts5_is_serialport_installed) {

		_ts5_is_serialport_installed = 1;

		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 serialport\n", 
			calling_function);

		// initialize event source
		al_init_user_event_source(
			&_ts5_data.timer.serialport_response_event_source);

		al_init_user_event_source(
			&_ts5_data.timer.serialport_trigger_event_source);

		// register event source
		al_register_event_source(_ts5_data.timer.response_queue,
			&_ts5_data.timer.serialport_response_event_source);
			
		al_register_event_source(_ts5_data.timer.trigger_queue,
			&_ts5_data.timer.serialport_trigger_event_source);

		al_register_event_source(_ts5_data.timer.trigger_log,
			&_ts5_data.timer.serialport_trigger_event_source);

		// initialize thread for serialport
		_ts5_data.timer.serialport_thread = 
			al_create_thread(ts5_serialport_threadfunc, NULL);
		
		al_start_thread(_ts5_data.timer.serialport_thread);

		atexit(ts5_uninstall_serialport);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Uninstall the serialport subsystem.
///
/// This function is called automatically at the end of the program.
////////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_serialport()
{
	if (_ts5_is_serialport_installed) {

		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 serialport\n");

		// remove thread
		al_join_thread(_ts5_data.timer.serialport_thread, NULL);
		al_destroy_thread(_ts5_data.timer.serialport_thread);

		// unregister event source
		al_unregister_event_source(_ts5_data.timer.response_queue,
			&_ts5_data.timer.serialport_response_event_source);

		al_unregister_event_source(_ts5_data.timer.trigger_queue,
			&_ts5_data.timer.serialport_trigger_event_source);

		al_unregister_event_source(_ts5_data.timer.trigger_log,
			&_ts5_data.timer.serialport_trigger_event_source);

		// remove event source
		al_destroy_user_event_source(
			&_ts5_data.timer.serialport_response_event_source);

		al_destroy_user_event_source(
			&_ts5_data.timer.serialport_trigger_event_source);

		// close ports
		int i;
		for (i=0; i<_ts5_status.timer.num_serialports; i++) {
		
			#ifdef TS5_WINDOWS
				if (!CloseHandle(*(_ts5_status.timer.serialport[i].port))) {
					ts5_log(TS5_LOGLEVEL_1,"%s %s %d\n", 
						"ts5_uninstall_serialport", 
						"could not uninstall serial port", i);
				}

			#else
				tcdrain(*(_ts5_status.timer.serialport[i].port));

				if (close(*(_ts5_status.timer.serialport[i].port)) == -1) {
					ts5_log(TS5_LOGLEVEL_1,"%s %s %d\n", 
						"ts5_uninstall_serialport", 
						"could not uninstall serial port", i);
				}
						
			#endif
			
			al_free(_ts5_status.timer.serialport[i].button_press_defined);
			_ts5_status.timer.serialport[i].button_press_defined = NULL;

			al_free(_ts5_status.timer.serialport[i].button_press_active);
			_ts5_status.timer.serialport[i].button_press_active = NULL;

			al_free(_ts5_status.timer.serialport[i].button_release_defined);
			_ts5_status.timer.serialport[i].button_release_defined = NULL;

			al_free(_ts5_status.timer.serialport[i].button_release_active);
			_ts5_status.timer.serialport[i].button_release_active = NULL;
			
			al_free(_ts5_status.timer.serialport[i].port);
		}
		
		al_free(_ts5_status.timer.serialport);
		_ts5_status.timer.serialport = NULL;
		_ts5_status.timer.num_serialports = 0;
		_ts5_status.timer.serialport_is_response_device = 0;
		_ts5_status.timer.serialport_is_trigger_device = 0;
	
		_ts5_is_serialport_installed = 0;
	}
}



////////////////////////////////////////////////////////////////////////////////
/// Serial port thread function.
///
/// \warning This is an internal function, no checks are performed.
////////////////////////////////////////////////////////////////////////////////
void *ts5_serialport_threadfunc()
{
	ALLEGRO_EVENT serialport_event;
	double previous_time;
	double current_time = al_get_time();

	#ifdef TS5_WINDOWS
	//	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL);
	
		// THREAD_PRIORITY_HIGHEST
		// THREAD_PRIORITY_ABOVE_NORMAL
		// THREAD_PRIORITY_NORMAL
		// THREAD_PRIORITY_TIME_CRITICAL
		
	#endif

	while (!al_get_thread_should_stop(_ts5_data.timer.serialport_thread)) {
		
		int i;
		for (i=0; i<_ts5_status.timer.num_serialports; i++) {


			// monitor responses and triggers
			if (_ts5_status.timer.serialport[i].num_defined_buttons ||
				_ts5_status.timer.serialport[i].is_trigger_input_device) {
				
				char serial_buff[TS5_MAX_CHAR];
				int num_bytes = 0;
				
				#ifdef TS5_WINDOWS
					unsigned long bytes_to_read = 1;
					unsigned long bytes_read = 0;
					
					if (!ReadFile(*(_ts5_status.timer.serialport[i].port), 
							serial_buff, bytes_to_read, &bytes_read, NULL)) {
						ts5_fatal("ts5_serialport_threadfunc: read error\n");
					}
					
					num_bytes = bytes_read;

				#else
					unsigned int bytes_to_read = 1;
					unsigned int bytes_read = 0;
					
					bytes_read = read(*(_ts5_status.timer.serialport[i].port), 
									serial_buff, bytes_to_read);

					if ((int)bytes_read == -1) {
						ts5_fatal("ts5_serialport_threadfunc: read error\n");
					}
				
					num_bytes = bytes_read;

				#endif
			
				if (num_bytes) {
										
					serialport_event.user.data1 = i;
					serialport_event.user.data2 = serial_buff[0];
					serialport_event.user.data3 = 
						(intptr_t)((current_time-previous_time)*1000000.0);				

					if (_ts5_status.timer.serialport[i].num_defined_buttons) {
						serialport_event.user.type = TS5_EVENT_SERIALPORT_BUTTON_DOWN;
						al_emit_user_event(
							&_ts5_data.timer.serialport_response_event_source,
							&serialport_event, NULL);	
					}
					
					if (_ts5_status.timer.serialport[i].is_trigger_input_device) {
					
						serialport_event.user.type = TS5_EVENT_SERIALPORT_TRIGGER;
						al_emit_user_event(
							&_ts5_data.timer.serialport_trigger_event_source,
							&serialport_event, NULL);
					}
					
				} 
			}

			// simulate triggers
			if (_ts5_status.timer.serialport[i].simulate_trigger_input) {

				if (current_time - _ts5_status.timer.serialport[i]
						.last_trigger_simulation
						>= _ts5_status.timer.serialport[i]
						.trigger_simulation_interval) {

					serialport_event.user.type = TS5_EVENT_SERIALPORT_TRIGGER;

					serialport_event.user.data1 = i;

					serialport_event.user.data2 = 
						_ts5_status.timer.serialport[i].simulate_trigger_input;

					serialport_event.user.data3 = 
						(intptr_t)((current_time-previous_time)*1000000.0);

					al_emit_user_event(
						&_ts5_data.timer.serialport_trigger_event_source,
						&serialport_event, NULL);

					_ts5_status.timer.serialport[i].last_trigger_simulation = 
						current_time;
				}
			}
	
		}
		previous_time = current_time;
		current_time = al_get_time();
		
		// wait 50 microseconds to avoid crashes on Windows 7
		do {
			
		} while (al_get_time() - current_time < 0.00005);
	}

	return NULL;
}


