////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file textio.c
/// Definitions of text input/output functions.
/// \example textio01.c
/// \example textio02.c
/// \example textio03.c
/// \example textio04.c
/// \example textio05.c
////////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/textio.h"
#include "../include/tscope5/display.h"
#include "../include/tscope5/graphics.h"
#include "../include/tscope5/textio_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/keyboard_internal.h"
#include "../include/tscope5/graphics_internal.h"


#include <allegro5/allegro_font.h>
#include <allegro5/allegro_ttf.h>


////////////////////////////////////////////////////////////////////////////////
/// @name Text input/output functions
/// The text input-output functions all take the smallest possible number of
/// parameters to put something on the screen: 
/// two coordinates and a format string.
/// 
/// The vertical coordinates are interpreted differently by 
/// the two coordinate systems. 
/// - When using the Cartesian coordinate system, 
///   text is vertically centered around the y coordinate. 
/// - When using the Display coordinate system, 
///   the y coordinate corresponds to the top of the text 
///   (like in other libraries).
///
/// The format string is the same as the standard C 
/// printf and scanf function families. 
/// 
/// Both standard C strings and Unicode strings can be used.
///
/// Drawing parameters are set with the graphics parameter functions.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Write text to the active bitmap.
///
/// \param x Horizontal position of the text.
/// \param y Vertical position of the text.
/// \param format Printf style format string.
///
/// \return Length of the string on the display (in pixels).
////////////////////////////////////////////////////////////////////////////////
double ts5_printf(double x, double y, const char *format, ...)
{
	ts5_check_textio("ts5_printf");
	ts5_log(TS5_LOGLEVEL_5, "ts5_printf(%f,%f,...)\n", x, y);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		x = ts5_relative_to_absolute_coordinate_x(x);
		y = ts5_relative_to_absolute_coordinate_y(y);
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {
		x = ts5_cartesian_to_display_coordinate_x(x);
		y = ts5_cartesian_to_display_coordinate_y(y 
				+ _ts5_status.graphics.font_size / 2.0);
	}

	x *= _ts5_status.display[_ts5_status.active_display].sampling_factor;
	y *= _ts5_status.display[_ts5_status.active_display].sampling_factor;

	TS5_USTR *buf;
	va_list ap;
	double width;

	va_start(ap, format);
	buf = al_ustr_new("");
	al_ustr_vappendf(buf, format, ap);
	va_end(ap);

	ALLEGRO_FONT *font;
	if (!_ts5_status.display[_ts5_status.active_display].multisampling) {
		font = _ts5_status.graphics.font->aliased_font;
	} 
	else {
		font = _ts5_status.graphics.font->multisampled_font;		
	} 
	
	al_draw_ustr(font, 
					_ts5_status.graphics.foreground_color,
					x, y, _ts5_status.graphics.text_alignment, buf);

	width = al_get_ustr_width(_ts5_status.graphics.font->aliased_font, buf);
	al_ustr_free(buf);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		width/=_ts5_status.graphics.target_width;
	}
	
	return width;
}


////////////////////////////////////////////////////////////////////////////////
/// Write justified text to the active bitmap.
///
/// \param x1 Horizontal start position of the text.
/// \param x2 Horizontal end position of the text.
/// \param y Vertical position of the text.
/// \param diff Maximum amount of space. If the amount of space in a line 
/// of text is greater than the diff value, left justified text will be written.
/// \param format Printf style format string.
////////////////////////////////////////////////////////////////////////////////
void ts5_printf_justify(double x1, double x2, double y, double diff, 
							const char *format, ...)
{
	ts5_check_textio("ts5_printf_justify");
	ts5_log(TS5_LOGLEVEL_5, "ts5_printf_justify(%f,%f,%f,%f,...)\n", 
				x1, x2, y, diff);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		x1 = ts5_relative_to_absolute_coordinate_x(x1);
		x2 = ts5_relative_to_absolute_coordinate_x(x2);
		y = ts5_relative_to_absolute_coordinate_y(y);
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {
		x1 = ts5_cartesian_to_display_coordinate_x(x1);
		x2 = ts5_cartesian_to_display_coordinate_x(x2);
		y = ts5_cartesian_to_display_coordinate_y(y 
				+ _ts5_status.graphics.font_size / 2.0);
	}

	x1 *= _ts5_status.display[_ts5_status.active_display].sampling_factor;
	x2 *= _ts5_status.display[_ts5_status.active_display].sampling_factor;
	y *= _ts5_status.display[_ts5_status.active_display].sampling_factor;
	diff *= _ts5_status.display[_ts5_status.active_display].sampling_factor;

	TS5_USTR *buf;
	va_list ap;

	va_start(ap, format);
	buf = al_ustr_new("");
	al_ustr_vappendf(buf, format, ap);
	va_end(ap);

	ALLEGRO_FONT *font;
	if (!_ts5_status.display[_ts5_status.active_display].multisampling) {
		font = _ts5_status.graphics.font->aliased_font;
	} 
	else {
		font = _ts5_status.graphics.font->multisampled_font;		
	} 
	
	al_draw_justified_ustr(font, 
							_ts5_status.graphics.foreground_color,
							x1, x2, y, diff, 
							_ts5_status.graphics.text_alignment, buf);
}


////////////////////////////////////////////////////////////////////////////////
/// Write UTF-8 encoded text to the active bitmap.
///
/// \param x Horizontal position of the text.
/// \param y Vertical position of the text.
/// \param ustr Unicode string.
///
/// \return Length of the string on the display (in pixels).
////////////////////////////////////////////////////////////////////////////////
double ts5_printf_ustr(double x, double y, const TS5_USTR *ustr)
{
	ts5_check_textio("ts5_printf_ustr");
	ts5_log(TS5_LOGLEVEL_5, "ts5_printf_ustr(%f,%f,...)\n", x, y);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		x = ts5_relative_to_absolute_coordinate_x(x);
		y = ts5_relative_to_absolute_coordinate_y(y);
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {
		x = ts5_cartesian_to_display_coordinate_x(x);
		y = ts5_cartesian_to_display_coordinate_y(y 
				+ _ts5_status.graphics.font_size / 2.0);
	}

	x *= _ts5_status.display[_ts5_status.active_display].sampling_factor;
	y *= _ts5_status.display[_ts5_status.active_display].sampling_factor;

	ALLEGRO_FONT *font;
	if (!_ts5_status.display[_ts5_status.active_display].multisampling) {
		font = _ts5_status.graphics.font->aliased_font;
	} 
	else {
		font = _ts5_status.graphics.font->multisampled_font;		
	} 
	
	al_draw_ustr(font, 
					_ts5_status.graphics.foreground_color,
					x, y, _ts5_status.graphics.text_alignment, ustr);

	double width;
	width = al_get_ustr_width(_ts5_status.graphics.font->aliased_font, ustr);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		width/=_ts5_status.graphics.target_width;
	}

	width /= _ts5_status.display[_ts5_status.active_display].sampling_factor;

	return width;
}


////////////////////////////////////////////////////////////////////////////////
/// Write justified UTF-8 encoded text to the active bitmap.
///
/// \param x1 Horizontal start position of the text.
/// \param x2 Horizontal end position of the text.
/// \param y Vertical position of the text.
/// \param diff Maximum amount of space. If the amount of space in a line 
/// of text is greater than the diff value, left justified text will be written.
/// \param ustr Unicode string.
////////////////////////////////////////////////////////////////////////////////
void ts5_printf_justify_ustr(double x1, double x2, double y, double diff, 
								const TS5_USTR *ustr)
{
	ts5_check_textio("ts5_printf_justify_ustr");
	ts5_log(TS5_LOGLEVEL_5, "ts5_printf_justify_ustr(%f,%f,%f,%f,...)\n", 
				x1, x2, y, diff);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		x1 = ts5_relative_to_absolute_coordinate_x(x1);
		x2 = ts5_relative_to_absolute_coordinate_x(x2);
		y = ts5_relative_to_absolute_coordinate_y(y);
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {
		x1 = ts5_cartesian_to_display_coordinate_x(x1);
		x2 = ts5_cartesian_to_display_coordinate_x(x2);
		y = ts5_cartesian_to_display_coordinate_y(y 
				+ _ts5_status.graphics.font_size / 2.0);
	}

	x1 *= _ts5_status.display[_ts5_status.active_display].sampling_factor;
	x2 *= _ts5_status.display[_ts5_status.active_display].sampling_factor;
	y *= _ts5_status.display[_ts5_status.active_display].sampling_factor;
	diff *= _ts5_status.display[_ts5_status.active_display].sampling_factor;

	ALLEGRO_FONT *font;
	if (!_ts5_status.display[_ts5_status.active_display].multisampling) {
		font = _ts5_status.graphics.font->aliased_font;
	} 
	else {
		font = _ts5_status.graphics.font->multisampled_font;		
	} 

	al_draw_justified_ustr(font, 
							_ts5_status.graphics.foreground_color,
							x1, x2, y, diff,
							_ts5_status.graphics.text_alignment, ustr);
}


////////////////////////////////////////////////////////////////////////////////
/// Read text from the keyboard.
///
/// \param x Horizontal position where the input will be echoed.
/// \param y Vertical position where the input will be echoed.
/// \param format Printf style format string.
///
/// \return The number of arguments that were processed.
////////////////////////////////////////////////////////////////////////////////
int ts5_scanf(double x, double y, const char *format, ...)
{
	ts5_check_textio("ts5_scanf");
	ts5_check_keyboard("ts5_scanf");
	ts5_log(TS5_LOGLEVEL_5, "ts5_scanf(%f,%f,...)\n", x, y);

	char str[TS5_MAX_CHAR];
	int done = 0, pos = 0;
	int ascii;
	TS5_COLOR fg = _ts5_status.graphics.foreground_color;
	TS5_COLOR bg = _ts5_status.graphics.background_color;

	ALLEGRO_EVENT_QUEUE *event_queue;
	event_queue = al_create_event_queue();

	if (!event_queue) {
		ts5_fatal("ts5_scanf: could not create event cue for keyboard\n");
		return 1;
	}

	al_register_event_source(event_queue, al_get_keyboard_event_source());

	ALLEGRO_EVENT event;

	do {
		al_wait_for_event(event_queue, &event);

		if (event.type == ALLEGRO_EVENT_KEY_CHAR) {

			ts5_set_foreground_color(bg);
			ts5_printf(x, y, "%s", str);

			ascii = event.keyboard.unichar;
			if (ascii == 3 || ascii == 13) {
				done = 1;
			}
			else if (ascii == 8 || ascii == 127) {
				if (pos > 0)
					pos--;
			}
			else if (ascii == 0) {

			}
			else {
				str[pos] = ascii;
				pos++;
			}

			if (pos > TS5_MAX_CHAR) {
				pos = TS5_MAX_CHAR;
			}
			str[pos] = '\0';
			ts5_set_foreground_color(fg);
			ts5_printf(x, y, "%s", str);
			ts5_flip_display();
		}
	} while (!done);

	al_unregister_event_source(event_queue, al_get_keyboard_event_source());
	al_destroy_event_queue(event_queue);

	int i;
	va_list args;
	va_start(args, format);
	i = vsscanf(str, format, args);
	va_end(args);

	return i;
}


////////////////////////////////////////////////////////////////////////////////
/// Get the width of a string without printing it.
///
/// \param format Printf style format string.
///
/// \return The width of the string on the display (in pixels).
////////////////////////////////////////////////////////////////////////////////
double ts5_get_text_width(const char *format, ...)
{
	ts5_check_textio("ts5_get_text_width");
	ts5_log(TS5_LOGLEVEL_5, "ts5_get_text_width(...)\n");

	TS5_USTR *buf;
	va_list ap;
	int width;

	va_start(ap, format);
	buf = al_ustr_new("");
	al_ustr_vappendf(buf, format, ap);
	va_end(ap);

	width = al_get_ustr_width(_ts5_status.graphics.font->aliased_font, buf);
	al_ustr_free(buf);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		width/=_ts5_status.graphics.target_width;
	}

	return width;
}


////////////////////////////////////////////////////////////////////////////////
/// Get the width of a UTF-8 string without printing it.
///
/// \param ustr Unicode string.
///
/// \return The width of the string on the display (in pixels).
////////////////////////////////////////////////////////////////////////////////
double ts5_get_text_width_ustr(TS5_USTR *ustr)
{
	ts5_check_textio("ts5_get_text_width_ustr");
	ts5_log(TS5_LOGLEVEL_5, "ts5_get_text_width_ustr(...)\n");

	double width;
	width = al_get_ustr_width(_ts5_status.graphics.font->aliased_font, ustr);

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		width/=_ts5_status.graphics.target_width;
	}

	return width;
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// @name Loading custom fonts
/// In Tscope5 three fonts are available as standard (courier, arial and times).
/// If these do not suit your needs any TrueType font that is installed 
/// on your computer can be loaded manually.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Load a user defined font.
///
/// \param fontfile Path to the file that describes the 1 type font.
/// \param size Size the font will be rendered in.
///
/// \return A pointer to the TS5_FONT font structure.
///
/// True type fonts can be found in /Library/fonts on Mac OS X.
////////////////////////////////////////////////////////////////////////////////
TS5_FONT *ts5_read_font(const char *fontfile, double size)
{
	ts5_check_textio("ts5_read_font");
	ts5_log(TS5_LOGLEVEL_4, "ts5_read_font(%s,%f)\n", fontfile, size);

	ts5_log(TS5_LOGLEVEL_2, "ts5_read_font: loading font %s %f\n", 
				fontfile, size);
	
	TS5_FONT *font;
	font = (TS5_FONT *)al_malloc(sizeof(TS5_FONT));
	font->aliased_font = al_load_ttf_font(fontfile, -size, 0);
	font->multisampled_font = al_load_ttf_font(fontfile, 
			-size * _ts5_status.display[_ts5_status.active_display].sampling_factor, 0);

	if (!font->aliased_font) {
		ts5_fatal("ts5_read_font: could not load font %s %d\n", fontfile,
					_ts5_status.graphics.font_size);
	}

	return font;
}


////////////////////////////////////////////////////////////////////////////////
/// Set a user defined font.
///
/// \param font The new font.
///
/// \return A pointer to the previous font.
////////////////////////////////////////////////////////////////////////////////
TS5_FONT *ts5_set_font(TS5_FONT *font)
{
	ts5_check_textio("ts5_set_font");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_font(%p)\n", font);

	if (font == NULL) {
		ts5_fatal("ts5_set_font: font is a NULL pointer\n");
	}

	TS5_FONT *retval = _ts5_status.graphics.font;
	_ts5_status.graphics.font = font;
	_ts5_status.graphics.font_index = TS5_USERFONT;

	_ts5_status.graphics.font_size = 
		al_get_font_line_height(_ts5_status.graphics.font->aliased_font);

	return retval;
}


////////////////////////////////////////////////////////////////////////////////
/// Free the memory used by a user defined font.
///
/// \param font pointer to the font that will be freed.
///
/// This function should be called at the end of the program for
/// each font read by the user.
////////////////////////////////////////////////////////////////////////////////
void ts5_free_font(TS5_FONT *font)
{
	ts5_check_textio("ts5_free_font");
	ts5_log(TS5_LOGLEVEL_2, "ts5_free_font(%p)\n", font);

	if (font) {
		al_destroy_font(font->aliased_font);
		al_destroy_font(font->multisampled_font);
		al_free(font);
		ts5_log(TS5_LOGLEVEL_2, "ts5_free_font: removed font\n");
		font = NULL;
	}
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////
