////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file textio_internal.c
/// Definitions of internal text input/output functions.
////////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/textio_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/bitmaps_internal.h"

#include <allegro5/allegro_font.h>
#include <allegro5/allegro_ttf.h>

/// Is the textio subsystem installed?
int _ts5_is_textio_installed = 0;


////////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each textio function.
///
/// \param calling_function Name the function that calls 
/// for this check or installation.
///
/// Checks whether the textio subsystem is installed. 
/// If not, the textio subsystem is installed.
///
/// Checks whether the drawing target is null. If null, the program is aborted.
///
/// Checks wheter a font is loaded. If not, a font is loaded.
///
/// The font files are looked for in the following places (in this order):
/// - the program directory
/// - the fonts subdirectory of the program directory
/// - /usr/local/share/tscope_data
////////////////////////////////////////////////////////////////////////////////
void ts5_check_textio(char *calling_function)
{
	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_textio\n", calling_function);

	if (!_ts5_is_textio_installed) {
		ts5_install_textio(calling_function);
	}

	if (!_ts5_data.target) {
		ts5_fatal("%s: no drawing target specified\n", calling_function);
	}

	if (_ts5_status.graphics.font_index == TS5_USERFONT) {

	}
	else if (!_ts5_status.graphics.fonts[_ts5_status.graphics.font_index]) {

		_ts5_status.graphics.fonts[_ts5_status.graphics.font_index] =
			(TS5_FONT *)al_malloc(sizeof(TS5_FONT));
		
		char path[TS5_MAX_CHAR];
		char path_with_prefix[TS5_MAX_CHAR];

		int factor = _ts5_status.display[_ts5_status.active_display].sampling_factor;
			
		if (_ts5_status.graphics.font_type == TS5_COURIER) {
			sprintf(path, "Courier New");
		}
		else if (_ts5_status.graphics.font_type == TS5_ARIAL) {
			sprintf(path, "Arial");
		}
		else if (_ts5_status.graphics.font_type == TS5_TIMES) {
			sprintf(path, "Times New Roman");
		}
		else {
			ts5_fatal("%s: invalid font type (%d)\n", calling_function,
						_ts5_status.graphics.font_type);
		}

		if (_ts5_status.graphics.font_style == TS5_REGULAR) {
		}
		else if (_ts5_status.graphics.font_style == TS5_BOLD) {
			sprintf(path, "%s Bold", path);
		}
		else if (_ts5_status.graphics.font_style == TS5_ITALIC) {
			sprintf(path, "%s Italic", path);
		}
		else if (_ts5_status.graphics.font_style == TS5_BOLD_ITALIC) {
			sprintf(path, "%s Bold Italic", path);
		}
		else {
			ts5_fatal("%s: invalid font style (%d)\n", calling_function,
						_ts5_status.graphics.font_style);
		}
		sprintf(path, "%s.ttf", path);

		// look in the program directory
		sprintf(path_with_prefix, "%s", path);

		_ts5_status.graphics.fonts[_ts5_status.graphics.font_index]->aliased_font =
			al_load_font(path_with_prefix, 
			_ts5_status.graphics.font_size, 0);

		_ts5_status.graphics.fonts[_ts5_status.graphics.font_index]->multisampled_font =
			al_load_font(path_with_prefix, 
			_ts5_status.graphics.font_size * factor, 0);

		// look in the fonts subdirectory
		if (!_ts5_status.graphics.fonts[_ts5_status.graphics.font_index]->aliased_font) {

			sprintf(path_with_prefix, "fonts/%s", path);

			_ts5_status.graphics.fonts[_ts5_status.graphics.font_index]->aliased_font =
				al_load_font(path_with_prefix, 
				_ts5_status.graphics.font_size, 0);

			_ts5_status.graphics.fonts[_ts5_status.graphics.font_index]->multisampled_font =
				al_load_font(path_with_prefix, 
				_ts5_status.graphics.font_size * factor, 0);
		}

		#ifndef TS5_WINDOWS
			
		// look in /usr/local/share/tscope5_data/
		if (!_ts5_status.graphics.fonts[_ts5_status.graphics.font_index]->aliased_font) {

			sprintf(path_with_prefix, "/usr/local/share/tscope5_data/%s", path);

			_ts5_status.graphics.fonts[_ts5_status.graphics.font_index]->aliased_font =
				al_load_font(path_with_prefix, 
				_ts5_status.graphics.font_size, 0);

			_ts5_status.graphics.fonts[_ts5_status.graphics.font_index]->multisampled_font =
				al_load_font(path_with_prefix, 
				_ts5_status.graphics.font_size * factor, 0);
		}

		#else
			
		// look in c:\mingw4tscope5\usr\local\share\tscope5_data
		if (!_ts5_status.graphics.fonts[_ts5_status.graphics.font_index]->aliased_font) {
		
			char *systemdrive = getenv("SYSTEMDRIVE");
			if (!systemdrive) {
				ts5_fatal("%s: could not determine system drive letter\n",
						calling_function);
			}	
			
			sprintf(path_with_prefix, "%s%s%s", systemdrive, 
				"/MinGW4Tscope5/msys/1.0/local/share/tscope5_data/", 
				path);
			
			_ts5_status.graphics.fonts[_ts5_status.graphics.font_index]->aliased_font =
				al_load_font(path_with_prefix, 
				_ts5_status.graphics.font_size, 0);

			_ts5_status.graphics.fonts[_ts5_status.graphics.font_index]->multisampled_font =
				al_load_font(path_with_prefix, 
				_ts5_status.graphics.font_size * factor, 0);
		}
		#endif
		

		ts5_log(TS5_LOGLEVEL_2, "%s: %s %s %d (index %d)\n",
					calling_function, "loading font", path, 
					_ts5_status.graphics.font_size,
					_ts5_status.graphics.font_index);

		_ts5_status.graphics.font =
			_ts5_status.graphics.fonts[_ts5_status.graphics.font_index];

		if (!_ts5_status.graphics.font) {
			ts5_fatal("%s: could not load font %s %d\n", calling_function, path,
						_ts5_status.graphics.font_size);
		}
	}
	else {
		_ts5_status.graphics.font =
			_ts5_status.graphics.fonts[_ts5_status.graphics.font_index];
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Install the textio subsystem.
///
/// \param calling_function Name the function that calls 
/// for this check or installation.
///
/// This function is called automatically if necessary.
////////////////////////////////////////////////////////////////////////////////
void ts5_install_textio(char *calling_function)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5(calling_function);
	}

	if (!_ts5_is_bitmaps_installed) {
		ts5_install_bitmaps(calling_function);
	}

	if (!_ts5_is_textio_installed) {

		_ts5_is_textio_installed = 1;
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 text input/output\n", 
			calling_function);

		al_init_font_addon();	// does not return succes/failure (yet)

		if (!al_init_ttf_addon()) {
			ts5_fatal("%s: could not install Tscope5 text input/output\n", 
				calling_function);
		}

		atexit(ts5_uninstall_textio);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Uninstall the textio subsystem.
///
/// This function is called automatically at the end of the program.
////////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_textio()
{
	if (_ts5_is_textio_installed) {

		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 text input/output\n");

		int i;
		for (i = 0; i < TS5_NFONTS; i++) {

			if (_ts5_status.graphics.fonts[i] != NULL) {

				#ifdef TS5_MACOSX
					al_destroy_font(_ts5_status.graphics.fonts[i]->aliased_font);
					al_destroy_font(_ts5_status.graphics.fonts[i]->multisampled_font);
					al_free(_ts5_status.graphics.fonts[i]);
					_ts5_status.graphics.fonts[i] = NULL;

				#endif

				ts5_log(TS5_LOGLEVEL_2, "%s: %s %d\n", 
					"Uninstalling Tscope5 text input/output", 
					"removed font", i);
			}
		}

		al_shutdown_ttf_addon();
		al_shutdown_font_addon();

		_ts5_is_textio_installed = 0;
	}
}
